/***************************************************************************
 *   Copyright (C) 2004 by Petric Frank                                    *
 *   pfrank@gmx.de                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/***************************************************************************
 *   This layer expects all file entries are stored by its checkout        *
 *   version, i.e filename@@1.5                                            *
 *   The example consists of two parts - a file name (= filename) and      *
 *   the version part (1.5). The version part is the version from where    *
 *   it was checked out from the repository.                               *
 *                                                                         *
 *   It is not allowed to have two files with the same file name part in   *
 *   a directory !                                                         *
 *   If it is necessary to check out an other version while an other       *
 *   version have been already checked out the user must either be warned  *
 *   before deleting the already existing version or this action mest be   *
 *   denied. In the latter case the user have to check in the file or      *
 *   un-checkout the file.                                                 *
 ***************************************************************************/

#include "fs-layer-checkout.h"
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/vfs.h>
#include <sys/xattr.h>
#include <dirent.h>
#include <utime.h>
#include <fcntl.h>
#include <unistd.h>
#include "cvs-repository.h"



static char *LayerAttrName	= "layer";
static char *LayerAttrValue	= "Checkout";



fs_layer_checkout::fs_layer_checkout (const std::string & base, cvs_repository *repo)
: fs_layer (base), repos (repo)
{
}



fs_layer_checkout::~fs_layer_checkout ()
{
  repos->decrement_refcount ();
}



directory *fs_layer_checkout::readdir (const std::string &path) const
{
  directory *dir = new directory ();
  std::string fullpath = cache_location + path;

  DIR *posix_dir = ::opendir (fullpath.c_str ());
  if (posix_dir != NULL)
  {
    struct dirent *entry;
    
    while ((entry = ::readdir (posix_dir)) != NULL)
    {
      std::string item = entry->d_name;
      std::string::size_type pos = item.find ("@@");

      // strip away version number
      if (pos != std::string::npos)
        item.erase (pos);
    
      dir->push_back (item);
    }
      
    ::closedir (posix_dir);
  }

  return dir;
}



int fs_layer_checkout::getattr (const std::string &path, struct stat *stbuf) const
{
  // The things become more complicated. Due the name given does not contain
  // the version requested i have to do a manual search.
  if (path.find ("@@") != std::string::npos)
    return fs_layer::getattr (path, stbuf);

  std::string realpath;
  if (!find_realname (path, &realpath))
    return -ENOENT;

  return fs_layer::getattr (realpath, stbuf);
}



int fs_layer_checkout::readlink (const std::string &path, char *buffer, size_t size) const
{
  // (symblic) links are not supported in this filesystem
  return -EINVAL;
}



int fs_layer_checkout::mknod (const std::string &path, mode_t mode, dev_t rdev)
{
  // allocation via filesystem is not allowed
  // must be done by VFS during check-out
  return -EINVAL;


  // must be a regular file - other node types are not supported
//  if (!S_ISREG (mode))
//    return -EINVAL;

//  std::string realpath;
  
//  if (!find_realname (path, &realpath))
//    return -ENOENT;

//  return fs_layer::mknod (realpath, mode, rdev);
}



int fs_layer_checkout::mkdir (const std::string &path, mode_t mode)
{
  // creation of subdirectories via filesystem is not allowed
  // must be done by VFS during check-out
  return -EINVAL;
  
//  std::string realpath;
  
//  if (!find_realname (path, &realpath))
//    return -ENOENT;

//  return fs_layer::mkdir (realpath, mode);
}



int fs_layer_checkout::unlink (const std::string &path)
{
  // this in fact removes the checked-out version and pops-up
  // the version from CVS repository
  if (path.find ("@@") != std::string::npos)
    return fs_layer::unlink (path);

  std::string realpath;
  
  if (!find_realname (path, &realpath))
    return -ENOENT;

  return fs_layer::unlink (realpath);
}



int fs_layer_checkout::rmdir (const std::string &path)
{
  if (path.find ("@@") != std::string::npos)
    return fs_layer::rmdir (path);

  std::string realpath;
  
  if (!find_realname (path, &realpath))
    return -ENOENT;

  return fs_layer::rmdir (realpath);
}



int fs_layer_checkout::symlink (const std::string &path, const std::string &dest)
{
  // (symblic) links are not supported in this filesystem
  return -EINVAL;
}



int fs_layer_checkout::rename (const std::string &path, const std::string &newpath)
{
  // rename is not supported in this filesystem
  // must be a rename to private layer
  // --> do it via VFS - command
  return -EINVAL;
}



int fs_layer_checkout::link (const std::string &path, const std::string &dest)
{
  // (symblic) links are not supported in this filesystem
  return -EINVAL;
}



int fs_layer_checkout::chmod (const std::string &path, mode_t mode)
{
  if (path.find ("@@") != std::string::npos)
    return fs_layer::chmod (path, mode);

  std::string realpath;
  
  if (!find_realname (path, &realpath))
    return -ENOENT;

  return fs_layer::chmod (realpath, mode);
}



int fs_layer_checkout::chown (const std::string &path, uid_t uid, gid_t gid)
{
  // ownership changes are not supported in this filesystem
  return -EINVAL;
}



int fs_layer_checkout::truncate (const std::string &path, off_t limit)
{
  if (path.find ("@@") != std::string::npos)
    return fs_layer::truncate (path, limit);

  std::string realpath;
  
  if (!find_realname (path, &realpath))
    return -ENOENT;

  return fs_layer::truncate (realpath, limit);
}



int fs_layer_checkout::utime (const std::string &path, struct utimbuf *buffer)
{
  if (path.find ("@@") != std::string::npos)
    return fs_layer::utime (path, buffer);

  std::string realpath;
  
  if (!find_realname (path, &realpath))
    return -ENOENT;

  return fs_layer::utime (realpath, buffer);
}



int fs_layer_checkout::open (const std::string &path, int flags)
{
  if (path.find ("@@") != std::string::npos)
    return fs_layer::open (path, flags);

  std::string realpath;

  if (!find_realname (path, &realpath))
    return -ENOENT;

  return fs_layer::open (realpath, flags);
}



int fs_layer_checkout::read (const std::string &path, char *buffer, size_t size, off_t offset) const
{
  if (path.find ("@@") != std::string::npos)
    return fs_layer::read (path, buffer, size, offset);

  std::string realpath;
  
  if (!find_realname (path, &realpath))
    return -ENOENT;

  return fs_layer::read (realpath, buffer, size, offset);
}



int fs_layer_checkout::write (const std::string &path, const char *buffer, size_t size, off_t offset)
{
  if (path.find ("@@") != std::string::npos)
    return fs_layer::write (path, buffer, size, offset);

  std::string realpath;
  
  if (!find_realname (path, &realpath))
    return -ENOENT;

  return fs_layer::write (realpath, buffer, size, offset);
}



int fs_layer_checkout::statfs (const std::string &path, struct statfs *stbuf) const
{
  if (path.find ("@@") != std::string::npos)
    return fs_layer::statfs (path, stbuf);

  std::string realpath;
  
  if (!find_realname (path, &realpath))
    return -ENOENT;

  return fs_layer::statfs (realpath, stbuf);
}



int fs_layer_checkout::setxattr (const std::string &path, const std::string &name, const std::string &value, size_t size, int flags)
{
  if (path.find ("@@") != std::string::npos)
    return fs_layer::setxattr (path, name.c_str (), value.c_str (), size, flags);

  std::string realpath;
  
  if (!find_realname (path, &realpath))
    return -EOPNOTSUPP;

  return fs_layer::setxattr (realpath, name.c_str (), value.c_str (), size, flags);
}



int fs_layer_checkout::getxattr (const std::string &path, const std::string &name, char *value, size_t size) const
{
  if (name == LayerAttrName)
  {
    if (size == 0)
      return strlen (LayerAttrValue) + 1;

    if (size <= strlen (LayerAttrValue))
      return -ERANGE;
      
    strcpy (value, LayerAttrValue);
    
    return strlen (LayerAttrValue) + 1;
  }

  std::string realpath;

  if (path.find ("@@") == std::string::npos)
  {  
    if (!find_realname (path, &realpath))
      return -EOPNOTSUPP;
  }
  else
    realpath = path;
  
  if (name == "checkin")
    return repos->checkin_file (realpath, value, size);

  if (name == "getlog")
    return repos->get_log (realpath, value, size);

  if (name == "version")
  {
    std::string::size_type pos = realpath.find ("@@");
    if (pos != std::string::npos)
      realpath.erase (0, pos + 2);
    else
      realpath.clear ();
    
    if (size == 0)
      return realpath.length () + 1;

    if (size <= realpath.length ())
      return -ERANGE;
	
    strcpy (value, realpath.c_str ());
    
    return realpath.length () + 1;
  }

  return fs_layer::getxattr (realpath, name.c_str (), value, size);
}



int fs_layer_checkout::listxattr (const std::string &path, char *list, size_t size) const
{
  if (path.find ("@@") != std::string::npos)
    return fs_layer::listxattr (path, list, size);

  std::string realpath;
  
  if (!find_realname (path, &realpath))
    return -EOPNOTSUPP;

  return fs_layer::listxattr (realpath, list, size);
}



int fs_layer_checkout::removexattr (const std::string &path, const std::string &name)
{
  if (path.find ("@@") != std::string::npos)
    return fs_layer::removexattr (path, name.c_str ());

  std::string realpath;
  
  if (!find_realname (path, &realpath))
    return -EOPNOTSUPP;

  return fs_layer::removexattr (realpath, name.c_str ());
}



bool fs_layer_checkout::find_realname (const std::string &path, std::string *realpath) const
{
  // The things become more complicated. Due the name given does not contain
  // the version requested i have to do a manual search.
  // In this filesystem i expect the non-versioned part of the filename is unique

  // split file name and path
  std::string name = cache_location + path;
  std::string::size_type pos = name.rfind ('/');

  // there must be a directory part
  if (pos == std::string::npos)
    return -EINVAL;
    
  std::string directory = name.substr (0, pos);
  name.erase (0, pos + 1);

  DIR *posix_dir = ::opendir (directory.c_str ());
  if (posix_dir != NULL)
  {
    struct dirent *entry;
    
    while ((entry = ::readdir (posix_dir)) != NULL)
    {
      std::string item = entry->d_name;
      std::string::size_type pos = item.find ("@@");

      // strip away version number
      if (pos != std::string::npos)
        item.erase (pos);

      // entry found - first match wins
      if (item == name)
      {
        // directory entry found - close directory search
        ::closedir (posix_dir);

        (*realpath).assign (directory + "/" + entry->d_name);
	(*realpath).erase (0, cache_location.length ());

        return true;
      }
    }
      
    ::closedir (posix_dir);
  }

  return false;
}
