/***************************************************************************
 *   Copyright (C) 2004 by Petric Frank                                    *
 *   pfrank@gmx.de                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "fs-layer.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/vfs.h>
#include <sys/xattr.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <utime.h>
#include <unistd.h>
// #include <pwd.h>



fs_layer::fs_layer (const std::string & base)
: cache_location (base)
{
  create_directory (cache_location, S_IRUSR | S_IWUSR | S_IXUSR);
}



fs_layer::~fs_layer ()
{
}



bool fs_layer::create_directory (const std::string &dir, mode_t mode)
{
  std::string path = dir;
  std::string::size_type pos;
  std::string dirpart;
  int dirmode = S_IFDIR | mode;

  if ((mode & S_IRUSR) != 0)
    dirmode |= S_IXUSR;

  if ((mode & S_IRGRP) != 0)
    dirmode |= S_IXGRP;

  if ((mode & S_IROTH) != 0)
    dirmode |= S_IXOTH;

  pos = 0;
  do
  {
    pos = path.find ('/', pos + 1);

    dirpart = path;
    if (pos != std::string::npos)
      dirpart.erase (pos);

    struct stat info;

    if (::lstat (dirpart.c_str (), &info) != 0)
    {
      int result = ::mkdir (dirpart.c_str (), dirmode);
      if (result != 0)
        return false;
    }
    else
      if (!S_ISDIR (info.st_mode))
        return false;

  } while (pos != std::string::npos);

  return true;
}



// *****************************************************
// helper methods - may be overridden by derived classes
// *****************************************************

repository *fs_layer::get_repository ()
{
  return 0;
}



int fs_layer::getattr (const std::string &path, struct stat *stbuf) const
{
  if (::lstat ((cache_location + path).c_str (), stbuf) < 0)
    return -errno;

  return 0;
}



int fs_layer::readlink (const std::string &path, char *buffer, size_t size) const
{
  int count = ::readlink ((cache_location + path).c_str (), buffer, size);
  
  if (count < 0)
    return -errno;

  buffer[count] = '\0';
  
  return 0;
}



int fs_layer::mknod (const std::string &path, mode_t mode, dev_t rdev)
{
  if (::mknod ((cache_location + path).c_str (), mode, rdev) < 0)
    return -errno;

  return 0;
}



int fs_layer::mkdir (const std::string &path, mode_t mode)
{
  if (::mkdir ((cache_location + path).c_str (), mode) < 0)
    return -errno;

  return 0;
}



int fs_layer::unlink (const std::string &path)
{
  if (::unlink ((cache_location + path).c_str ()) < 0)
    return -errno;

  return 0;
}



int fs_layer::rmdir (const std::string &path)
{
  if (::rmdir ((cache_location + path).c_str ()) < 0)
    return -errno;

  return 0;
}



int fs_layer::symlink (const std::string &path, const std::string &dest)
{
  if (::symlink ((cache_location + path).c_str (), (cache_location + dest).c_str ()) < 0)
    return -errno;

  return 0;
}



int fs_layer::rename (const std::string &path, const std::string &newpath)
{
  if (::rename ((cache_location + path).c_str (), (cache_location + newpath).c_str ()) < 0)
    return -errno;

  return 0;
}



int fs_layer::link (const std::string &path, const std::string &dest)
{
  if (::link ((cache_location + path).c_str (), (cache_location + dest).c_str ()) < 0)
    return -errno;

  return 0;
}



int fs_layer::chmod (const std::string &path, mode_t mode)
{
  if (::chmod ((cache_location + path).c_str (), mode) < 0)
    return -errno;

  return 0;
}



int fs_layer::chown (const std::string &path, uid_t uid, gid_t gid)
{
  if (::lchown ((cache_location + path).c_str (), uid, gid) < 0)
    return -errno;

  return 0;
}



int fs_layer::truncate (const std::string &path, off_t limit)
{
  if (::truncate ((cache_location + path).c_str (), limit) < 0)
    return -errno;

  return 0;
}



int fs_layer::utime (const std::string &path, struct utimbuf *buffer)
{
  if (::utime ((cache_location + path).c_str (), buffer) < 0)
    return -errno;

  return 0;
}



int fs_layer::open (const std::string &path, int flags)
{
  int handle = ::open ((cache_location + path).c_str (), flags);

  if (handle < 0)
    return -errno;
  
  close (handle);

  return 0;
}



int fs_layer::read (const std::string &path, char *buffer, size_t size, off_t offset) const
{
  int result;
  int fin = ::open ((cache_location + path).c_str (), O_RDONLY);
  
  if (fin < 0)
    return -errno;
  
  if ((result = pread (fin, buffer, size, offset)) < 0)
    result = -errno;

  close (fin);
  
  return result;
}



int fs_layer::write (const std::string &path, const char *buffer, size_t size, off_t offset)
{
  int result;
  int fout = ::open ((cache_location + path).c_str (), O_WRONLY);
  
  if (fout < 0)
    return -errno;
  
  if ((result = pwrite (fout, buffer, size, offset)) < 0)
    result = -errno;

  close (fout);
  
  return result;
}



int fs_layer::statfs (const std::string &path, struct statfs *stbuf) const
{
  if (::statfs ((cache_location + path).c_str (), stbuf) < 0)
    return -errno;

  return 0;
}



int fs_layer::flush (const std::string &path)
{
  return 0;
}



int fs_layer::release (const std::string &path, int flags)
{
  return 0;
}



int fs_layer::fsync (const std::string &path, int isdatasync)
{
  return 0;
}



int fs_layer::setxattr (const std::string &path, const std::string &name, const std::string &value, size_t size, int flags)
{
  if (::lsetxattr ((cache_location + path).c_str (), name.c_str (), value.c_str (), size, flags) < 0)
    return -errno;

  return 0;
}



int fs_layer::getxattr (const std::string &path, const std::string &name, char *value, size_t size) const
{
  if (::lgetxattr ((cache_location + path).c_str (), name.c_str (), value, size) < 0)
    return -errno;

  return 0;
}



int fs_layer::listxattr (const std::string &path, char *list, size_t size) const
{
  int result = ::llistxattr ((cache_location + path).c_str (), list, size);

  if (result < 0)
    return -errno;

  return result;
}



int fs_layer::removexattr (const std::string &path, const std::string &name)
{
  if (::lremovexattr ((cache_location + path).c_str (), name.c_str ()) < 0)
    return -errno;

  return 0;
}
