/***************************************************************************
 *   Copyright (C) 2004 by Petric Frank                                    *
 *   pfrank@gmx.de                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "view.h"
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <sstream>
#include "fs-layer.h"
#include "repository.h"
#include "system-log.h"



view::view ()
: PrivateLayer (0)
{
}



view::~view ()
{
  purge_layers ();
}



void view::shutdown ()
{
  ListType::const_iterator iter;

  for (iter = layers.begin (); iter != layers.end (); ++iter)
  {
    char buffer[256];

    if ((*iter)->getxattr ("/", "layer", buffer, 256) >= 0)
      if (strcmp (buffer, "Checkout") != 0)
      {
        repository *repo = (*iter)->get_repository ();

	// let every repository save its state
	if (repo)
	  repo->shutdown ();
      }
  }
}



void view::set_private_layer (fs_layer *layer)
{
  PrivateLayer = layer;
}



void view::purge_layers ()
{
  ListType::const_iterator iter;
  
  for (iter = layers.begin (); iter != layers.end (); ++iter)
    delete (*iter);

  layers.clear ();
}



void view::append_layer (fs_layer *layer)
{
  layers.push_back (layer);
}



int view::add_item (const std::string &path, char *response, size_t maxsize)
{
  struct stat stbuf;

  if (PrivateLayer->getattr (path, &stbuf) < 0)
  {
    std::ostringstream errmsg;
    system_log *log = system_log::instance ();

    errmsg << "'" << path << "' is not a private file.";

    log->error << errmsg.str () << std::endl;

    return handle_response (errmsg.str (), response, maxsize);
  }

  ListType::const_iterator iter;

  for (iter = layers.begin (); iter != layers.end (); ++iter)
  {
    repository *repo = (*iter)->get_repository ();
    
    if (repo)
      return repo->add_item (path, response, maxsize, PrivateLayer);
  }

  // adding failed - no CVS layer seems to be able to add the item
  return -EOPNOTSUPP;
}



directory *view::readdir (const std::string &path) const
{
  directory *dir = new directory ();
  directory *part;

  ListType::const_iterator iter;
  
  for (iter = layers.begin (); iter != layers.end (); ++iter)
  {
    part = (*iter)->readdir (path);
    
    part->sort ();		// sort returned list
    
    dir->merge (*part);		// sort contents of 'part' into 'dir'
    
    dir->unique ();		// remove duplicate entries
    
    delete part;		// dispose remaining empty list
  }

  if (PrivateLayer)
  {
    part = PrivateLayer->readdir (path);
    
    part->sort ();		// sort returned list
    
    dir->merge (*part);		// sort contents of 'part' into 'dir'
    
    dir->unique ();		// remove duplicate entries
    
    delete part;		// dispose remaining empty list
  }

  return dir;			// pass ownership to caller !
}



int view::getattr (const std::string &path, struct stat *stbuf) const
{
  ListType::const_iterator iter;

  for (iter = layers.begin (); iter != layers.end (); ++iter)
  {
    if ((*iter)->getattr (path, stbuf) >= 0)
      return 0;			// first match wins
  }

  // at least check the list of private files
  if (PrivateLayer)
    return PrivateLayer->getattr (path, stbuf);

  return -EIO;
}



int view::readlink (const std::string &path, char *buffer, size_t size) const
{
  ListType::const_iterator iter;
  int count = -EIO;
  
  for (iter = layers.begin (); iter != layers.end (); ++iter)
  {
    count = (*iter)->readlink (path, buffer, size);
    
    if (count >= 0)
      return count;		// first match wins
  }

  // at least check the list of private files
  if (PrivateLayer)
    return PrivateLayer->readlink (path, buffer, size);

  return count;
}



int view::mknod (const std::string &path, mode_t mode, dev_t rdev)
{
  ListType::iterator iter;
  int result = -EIO;
  
  for (iter = layers.begin (); iter != layers.end (); ++iter)
  {
    result = (*iter)->mknod (path, mode, rdev);
    
    if (result >= 0)
      return result;		// first successful operation wins
  }

  // at least check the list of private files
  if (PrivateLayer)
    return PrivateLayer->mknod (path, mode, rdev);

  return result;
}



int view::mkdir (const std::string &path, mode_t mode)
{
  ListType::iterator iter;
  int result = -EIO;
  
  for (iter = layers.begin (); iter != layers.end (); ++iter)
  {
    result = (*iter)->mkdir (path, mode);
    
    if (result >= 0)
      return result;		// first successful operation wins
  }

  // at least check the list of private files
  if (PrivateLayer)
    return PrivateLayer->mkdir (path, mode);

  return result;
}



int view::unlink (const std::string &path)
{
  ListType::iterator iter;
  int result = -EIO;
  
  for (iter = layers.begin (); iter != layers.end (); ++iter)
  {
    result = (*iter)->unlink (path);
    
    if (result >= 0)
      return result;		// first match wins
  }

  // at least check the list of private files
  if (PrivateLayer)
    return PrivateLayer->unlink (path);

  return result;
}



int view::rmdir (const std::string &path)
{
  ListType::iterator iter;
  int result = -EIO;
  
  for (iter = layers.begin (); iter != layers.end (); ++iter)
  {
    result = (*iter)->rmdir (path);
    
    if (result >= 0)
      return result;		// first match wins
  }

  // at least check the list of private files
  if (PrivateLayer)
    return PrivateLayer->rmdir (path);

  return result;
}



int view::symlink (const std::string &path, const std::string &dest)
{
  ListType::iterator iter;
  int result = -EIO;
  
  for (iter = layers.begin (); iter != layers.end (); ++iter)
  {
    result = (*iter)->symlink (path, dest);
    
    if (result >= 0)
      return result;		// first match wins
  }

  // at least check the list of private files
  if (PrivateLayer)
    return PrivateLayer->symlink (path, dest);

  return result;
}



int view::rename (const std::string &path, const std::string &newpath)
{
  ListType::iterator iter;
  int result = -EIO;
  
  for (iter = layers.begin (); iter != layers.end (); ++iter)
  {
    result = (*iter)->rename (path, newpath);
    
    if (result >= 0)
      return result;		// first match wins
  }

  // at least check the list of private files
  if (PrivateLayer)
    return PrivateLayer->rename (path, newpath);

  return result;
}



int view::link (const std::string &path, const std::string &dest)
{
  ListType::iterator iter;
  int result = -EIO;
  
  for (iter = layers.begin (); iter != layers.end (); ++iter)
  {
    result = (*iter)->link (path, dest);
    
    if (result >= 0)
      return result;		// first match wins
  }

  // at least check the list of private files
  if (PrivateLayer)
    return PrivateLayer->link (path, dest);

  return result;
}



int view::chmod (const std::string &path, mode_t mode)
{
  ListType::iterator iter;
  int result = -EIO;
  
  for (iter = layers.begin (); iter != layers.end (); ++iter)
  {
    result = (*iter)->chmod (path, mode);
    
    if (result >= 0)
      return result;		// first match wins
  }

  // at least check the list of private files
  if (PrivateLayer)
    return PrivateLayer->chmod (path, mode);

  return result;
}



int view::chown (const std::string &path, uid_t uid, gid_t gid)
{
  ListType::iterator iter;
  int result = -EIO;
  
  for (iter = layers.begin (); iter != layers.end (); ++iter)
  {
    result = (*iter)->chown (path, uid, gid);
    
    if (result >= 0)
      return result;		// first match wins
  }

  // at least check the list of private files
  if (PrivateLayer)
    return PrivateLayer->chown (path, uid, gid);

  return result;
}



int view::truncate (const std::string &path, off_t limit)
{
  ListType::iterator iter;
  int result = -EIO;
  
  for (iter = layers.begin (); iter != layers.end (); ++iter)
  {
    result = (*iter)->truncate (path, limit);
    
    if (result >= 0)
      return result;		// first match wins
  }

  // at least check the list of private files
  if (PrivateLayer)
    return PrivateLayer->truncate (path, limit);

  return result;
}



int view::utime (const std::string &path, struct utimbuf *buffer)
{
  ListType::iterator iter;
  int result = -EIO;
  
  for (iter = layers.begin (); iter != layers.end (); ++iter)
  {
    result = (*iter)->utime (path, buffer);
    
    if (result >= 0)
      return result;		// first match wins
  }

  // at least check the list of private files
  if (PrivateLayer)
    return PrivateLayer->utime (path, buffer);

  return result;
}



int view::open (const std::string &path, int flags)
{
  ListType::iterator iter;
  int result = -EIO;

  for (iter = layers.begin (); iter != layers.end (); ++iter)
  {
    result = (*iter)->open (path, flags);
    
    if (result >= 0)
      return result;		// first match wins
  }

  // at least check the list of private files
  if (PrivateLayer)
    return PrivateLayer->open (path, flags);

  return result;
}



int view::read (const std::string &path, char *buffer, size_t size, off_t offset) const
{
  ListType::const_iterator iter;
  int count = -EIO;
  
  for (iter = layers.begin (); iter != layers.end (); ++iter)
  {
    count = (*iter)->read (path, buffer, size, offset);
    
    if (count >= 0)
      return count;		// first match wins
  }

  // at least check the list of private files
  if (PrivateLayer)
    return PrivateLayer->read (path, buffer, size, offset);

  return count;
}



int view::write (const std::string &path, const char *buffer, size_t size, off_t offset)
{
  ListType::iterator iter;
  int count = -EIO;
  
  for (iter = layers.begin (); iter != layers.end (); ++iter)
  {
    count = (*iter)->write (path, buffer, size, offset);
    
    if (count >= 0)
      return count;		// first match wins
  }

  // at least check the list of private files
  if (PrivateLayer)
    return PrivateLayer->write (path, buffer, size, offset);

  return count;
}



int view::statfs (const std::string &path, struct statfs *stbuf) const
{
  ListType::const_iterator iter;
  int count = -EIO;
  
  for (iter = layers.begin (); iter != layers.end (); ++iter)
  {
    count = (*iter)->statfs (path, stbuf);
    
    if (count >= 0)
      return count;		// first match wins
  }

  // at least check the list of private files
  if (PrivateLayer)
    return PrivateLayer->statfs (path, stbuf);

  return count;
}



int view::flush (const std::string &path)
{
  ListType::iterator iter;
  int count = -EIO;
  
  for (iter = layers.begin (); iter != layers.end (); ++iter)
  {
    count = (*iter)->flush (path);
    
    if (count >= 0)
      return count;		// first match wins
  }

  // at least check the list of private files
  if (PrivateLayer)
    return PrivateLayer->flush (path);

  return count;
}



int view::release (const std::string &path, int flags)
{
  ListType::iterator iter;
  int count = -EIO;
  
  for (iter = layers.begin (); iter != layers.end (); ++iter)
  {
    count = (*iter)->release (path, flags);
    
    if (count >= 0)
      return count;		// first match wins
  }

  // at least check the list of private files
  if (PrivateLayer)
    return PrivateLayer->release (path, flags);

  return count;
}



int view::fsync (const std::string &path, int fdatasync)
{
  ListType::iterator iter;
  int count = -EIO;
  
  for (iter = layers.begin (); iter != layers.end (); ++iter)
  {
    count = (*iter)->fsync (path, fdatasync);
    
    if (count >= 0)
      return count;		// first match wins
  }

  // at least check the list of private files
  if (PrivateLayer)
    return PrivateLayer->fsync (path, fdatasync);

  return count;
}



int view::setxattr (const std::string &path, const std::string &name, const std::string &value, size_t size, int flags)
{
  ListType::iterator iter;
  int count = -EIO;
  
  for (iter = layers.begin (); iter != layers.end (); ++iter)
  {
    count = (*iter)->setxattr (path, name, value, size, flags);
    
    if (count >= 0)
      return count;		// first match wins
  }

  // at least check the list of private files
  if (PrivateLayer)
    return PrivateLayer->setxattr (path, name, value, size, flags);

  return count;
}



int view::getxattr (const std::string &path, const std::string &name, char *value, size_t size) const
{
  ListType::const_iterator iter;
  int count = -EIO;
  
  for (iter = layers.begin (); iter != layers.end (); ++iter)
  {
    struct stat stbuf;

    if ((*iter)->getattr (path, &stbuf) >= 0)
      count = (*iter)->getxattr (path, name, value, size);

    if (count >= 0)
      return count;		// first match wins
  }

  // at least check the list of private files
  if (PrivateLayer)
  {
    if (name == "add")
      return const_cast<view *> (this)->add_item (path, value, size);

    return PrivateLayer->getxattr (path, name, value, size);
  }

  return count;
}



int view::listxattr (const std::string &path, char *list, size_t size) const
{
  ListType::const_iterator iter;
  int count = -EIO;
  
  for (iter = layers.begin (); iter != layers.end (); ++iter)
  {
    count = (*iter)->listxattr (path, list, size);
    
    if (count >= 0)
      return count;		// first match wins
  }

  // at least check the list of private files
  if (PrivateLayer)
    return PrivateLayer->listxattr (path, list, size);

  return count;
}



int view::removexattr (const std::string &path, const std::string &name)
{
  ListType::iterator iter;
  int count = -EIO;
  
  for (iter = layers.begin (); iter != layers.end (); ++iter)
  {
    count = (*iter)->removexattr (path, name);
    
    if (count >= 0)
      return count;		// first match wins
  }

  // at least check the list of private files
  if (PrivateLayer)
    return PrivateLayer->removexattr (path, name);

  return count;
}



int view::handle_response (const std::string &response, char *buffer, size_t maxsize)
{
  if (maxsize != 0)
  {
    if (maxsize < response.length ())
      return -ERANGE;

    strcpy (buffer, response.c_str ());
  }

  return response.length () + 1;
}
